<?php

namespace SGalinski\Utility;

/*******************************************************************************
 *
 * Copyright (c) sgalinski Internet Services (http://www.sgalinski.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 ******************************************************************************/

use Gitlab\Api\Groups;
use Gitlab\Api\Projects;
use Gitlab\Api\Users;
use Gitlab\Client;
use Gitlab\Exception\RuntimeException;

/**
 * Class with helper methods for the gitlab webhook.
 */
class GeneralUtility {
	/**
	 * @var array
	 */
	protected $settings = [];

	/**
	 * @var Client
	 */
	protected $client;

	/**
	 * Construct for general utility
	 */
	public function __construct($settings, Client $client = NULL) {
		$this->setSettings($settings);

		if (!$client) {
			$client = new Client($this->settings['Configuration']['Gitlab']['apiUrl']);
			$client->authenticate($this->settings['Configuration']['Gitlab']['authToken'], Client::AUTH_URL_TOKEN);
		}
		$this->setClient($client);
	}

	/**
	 * Method gets a commit message as string and the id of the project. With the project
	 * id we can get the web url to the issue and if we find issue number with #number in the
	 * commit message, we replace the number with a link to the issue.
	 *
	 * @param String $commitMessage
	 * @param array $projectData
	 * @return String
	 */
	public function findIssueIdsAndLinkThem($commitMessage, $projectData) {
		if (!count($projectData)) {
			return $commitMessage;
		}

		$processingCommitMessage = $commitMessage;
		preg_match_all('/#[0-9]{1,4}/', $processingCommitMessage, $matches, PREG_SET_ORDER);

		// get url of an issue
		$issueUrl = '';
		if (count($matches)) {
			$issueUrl = $projectData['web_url'];
		}

		// generate the markup of a link to the issue and replace the #issue in the commit message
		foreach ($matches as $entry) {
			if (strlen($issueUrl)) {
				$entryUrl = $issueUrl . '/issues/' . str_replace('#', '', $entry[0]);
				$entryLink = '<a href="' . $entryUrl . '" target="_blank">' . $entry[0] . '</a>';
				$commitMessage = str_replace($entry[0], $entryLink, $commitMessage);
			}
		}

		return $commitMessage;
	}

	/**
	 * Method checks the commit log if the given commit id is known.
	 *
	 * @param string $commitId
	 * @return boolean
	 */
	public function checkLogForCommitId($commitId) {
		$commitFailures = file_get_contents($this->settings['Configuration']['Logging']['commitIdLog']);
		if ($commitFailures === FALSE) {
			$this->createLogFile();
		}

		return (strpos($commitFailures, $commitId) > 0) ? TRUE : FALSE;
	}

	/**
	 * Actually we need to log the commit ids, because there is an error in gitlab. Sometimes the
	 * webhook is triggered twice and so we send the mails twice. This method logs the commit id.
	 *
	 * @param string $commitId
	 * @return void
	 */
	public function logCommitId($commitId) {
		// need double ticks here because the line break does not work with single ticks
		$commitIdLog = $commitId . ",\r\n";
		$logFilePath = $this->settings['Configuration']['Logging']['commitIdLog'];

		if (file_put_contents($logFilePath, $commitIdLog, FILE_APPEND | LOCK_EX) === FALSE) {
			$this->createLogFile();
			file_put_contents($logFilePath, $commitIdLog, FILE_APPEND | LOCK_EX);
		}
	}

	/**
	 * Create a log file for the commit ids.
	 *
	 * @return void
	 */
	protected function createLogFile() {
		$filename = $this->settings['Configuration']['Logging']['commitIdLog'];
		// need double ticks here because the line break does not work with single ticks
		$firstLine = "# Log file for commit ids.\r\n";

		$file = fopen($filename, 'a');
		if (!$file) {
			error_log('Cannot open file (' . $filename . ')');
			exit;
		}

		if (fwrite($file, $firstLine) === FALSE) {
			error_log('Cannot write to file (' . $filename . ')');
			exit;
		}

		fclose($file);
	}

	/**
	 * The method fetch the project members and the members of the group. The name and email address
	 * of the fetched persons will be returned in an array.
	 *
	 * @param array $projectData
	 * @return array
	 */
	public function getRecipientsForProject(array $projectData) {
		if (!count($projectData)) {
			return [];
		}
		$users = [];
		$recipients = [];
		$fallbackRecipient = [$this->settings['Configuration']['MailConfiguration']['fallbackRecipient']];
		$groupId = intval($projectData['namespace']['id']);

		// get members of the project
		$projectMember = $this->getProjectMembers($projectData['id']);
		foreach ($projectMember as $member) {
			$users[] = $member['id'];
		}

		$group = $this->getMemberDataOfAGroup($groupId);
		foreach ($group as $member) {
			$users[] = $member['id'];
		}

		$users = array_unique($users);
		foreach ($users as $user) {
			$userData = $this->getUser($user);
			if ($userData['state'] === 'active') {
				$recipients[$userData['email']] = $userData['name'];
			}
		}
		return (!count($recipients)) ? $fallbackRecipient : $recipients;
	}

	/**
	 * Get member for a given project id. Returns an empty array if there are no
	 * members for the given project.
	 *
	 * @param integer $id
	 * @return array
	 */
	public function getProjectMembers($id) {
		$members = [];

		try {
			/** @var Projects $projects */
			$projects = $this->client->projects();
			/** @var array $members */
			$members = $projects->members($id);
		} catch (RuntimeException $e) {
			error_log('Unable to get project member data' . $e->getMessage());
		}

		return count($members) ? $members : [];
	}

	/**
	 * Get the gitlab users. Returns an empty array if no users can be found.
	 *
	 * @return array
	 */
	public function getUsers() {
		$userData = [];

		try {
			/** @var Users $users */
			$users = $this->client->users();
			/**
			 * Since commit 039883d the gitlab api php framework
			 * uses pagination. And there is no flag to fetch all. So
			 * we first use a high value to fetch "all".
			 *
			 * @var array $userData
			 */
			$userData = $users->all(NULL, 1, 9999);
		} catch (RuntimeException $e) {
			error_log('Unable to get user data' . $e->getMessage());
		}

		return count($userData) ? $userData : [];
	}

	/**
	 * Get a gitlab user
	 *
	 * @param int $id 
	 * @return array
	 */
	public function getUser($id) {
		$userData = [];

		try {
			/** @var Users $users */
			$users = $this->client->users();
			$userData = $users->show($id);
		} catch (RuntimeException $e) {
			error_log('Unable to get user data:' . $e->getMessage());
		}

		return count($userData) ? $userData : [];
	}

	/**
	 * Get mail addresses from a gitlab user
	 *
	 * @param int $id 
	 * @return array
	 */
	public function getUserEmails($id) {
		$userData = [];

		try {
			/** @var Users $users */
			$users = $this->client->users();
			$userData = $users->emails($id);
		} catch (RuntimeException $e) {
			error_log('Unable to get user email: ' . $e->getMessage());
		}

		return count($userData) ? $userData : [];
	}

	/**
	 * Method gets a id of a group and returns an array with the member data of the group.
	 * If we can't find the group we return NULL.
	 *
	 * @param Integer $groupId
	 * @return array
	 */
	public function getMemberDataOfAGroup($groupId) {
		$groupData = [];

		try {
			/** @var Groups $groups */
			$groups = $this->client->groups();
			/** @var array $groupData */
			$groupData = $groups->members($groupId);
		} catch (RuntimeException $e) {
			error_log('Unable to get group data: ' . $e->getMessage());
		}

		return count($groupData) ? $groupData : [];
	}

	/**
	 * Setter for settings.
	 *
	 * @param array $settings
	 * @return void
	 */
	public function setSettings($settings) {
		$this->settings = $settings;
	}

	/**
	 * Getter for debug.
	 *
	 * @return array
	 */
	public function getSettings() {
		return $this->settings;
	}

	/**
	 * Getter for client
	 *
	 * @return Client
	 */
	public function getClient() {
		return $this->client;
	}

	/**
	 * Setter for client
	 *
	 * @param Client $client
	 * @return void
	 */
	public function setClient(Client $client) {
		$this->client = $client;
	}
}
