<?php

namespace SGalinski;

/*******************************************************************************
 *
 * Copyright (c) sgalinski Internet Services (http://www.sgalinski.de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 ******************************************************************************/

use Gitlab\Api\Projects;
use Gitlab\Client;
use Gitlab\Exception\RuntimeException;
use Gitlab\HttpClient\Listener\AuthListener;
use SGalinski\Utility\GeneralUtility;
use Symfony\Component\Mailer\Transport;
use Symfony\Component\Mailer\Transport\SendmailTransport;
use Symfony\Component\Mailer\Mailer;
use Symfony\Component\Mime\Email;
use Symfony\Component\DomCrawler\Crawler;
use Symfony\Component\Mime\Address;
use TijsVerkoyen\CssToInlineStyles\CssToInlineStyles;
use Twig_Environment;
use Twig_Loader_Filesystem;
use Twig_Template;

/**
 * The web hook class is responsible for the web hook processing.
 */
class Webhook {
	/**
	 * @var array
	 */
	protected $settings = [];

	/**
	 * Array with the json data of the push to the repository.
	 *
	 * @var array
	 */
	protected $request = [];

	/**
	 * @var Twig_Template
	 */
	protected $view;

	/**
	 * @var boolean
	 */
	protected $debugMode = FALSE;

	/**
	 * @var array
	 */
	static protected $diffCache = [];

	/**
	 * @var \Gitlab\Client
	 */
	protected $apiClient;

	/**
	 * Construct sets request and setting property for the web hook object.
	 *
	 * @param $jsonString
	 * @param array $settings
	 */
	public function __construct($jsonString, array $settings) {
		$this->setRequest(json_decode($jsonString, TRUE));
		$this->setSettings($settings);

		// set up view
		$loader = new Twig_Loader_Filesystem('Resources/Templates');
		$twig = new Twig_Environment($loader, ['cache' => 'Resources/TemplatesCached']);
		$this->view = $twig->loadTemplate('Commit.html');

		// set debug mode
		$this->debugMode = ($this->settings['Configuration']['Debug']['enableDebugMode'] === TRUE) ? TRUE : FALSE;
	}

	public function processWebhook() {
		$configuration = $this->settings['Configuration']['Gitlab'];
		$authenticationToken = ($configuration['authToken'] ? $configuration['authToken'] : NULL);
		$sudo = ($configuration['sudo'] ? $configuration['sudo'] : NULL);

		$this->apiClient = $apiClient = new \Gitlab\Client();
		$this->apiClient->setUrl($this->settings['Configuration']['Gitlab']['apiUrl']);
		$this->apiClient->authenticate($authenticationToken, \Gitlab\Client::AUTH_HTTP_TOKEN, $sudo);
		$utility = new GeneralUtility($this->settings, $apiClient);

		$branchName = str_replace('refs/heads/', '', $this->request['ref']);
		$projectId = $this->request['project_id'];

		#error_log(print_r($this->request, TRUE));

		try {
			/** @var Projects $projects */
			$projects = $apiClient->projects();
			/** @var array $members */
			$projectData = $projects->show($projectId);
		} catch (RuntimeException $e) {
			error_log(sprintf('Unable to get project data', $e->getMessage()));
			exit;
		}

		foreach ($this->request['commits'] as $commit) {
			// skip commit if the commit has been logged
			if ($utility->checkLogForCommitId($commit['id'])) {
				continue;
			}

			// adjust commit message
			$commit['message'] = $utility->findIssueIdsAndLinkThem($commit['message'], $projectData);

			$bindings = [
				'username' => $this->request['user_name'],
				'branch' => $this->request['ref'],
				'repository' => $projectData['name_with_namespace'],
				'commit' => $commit,
				'diff' => $this->crawlCommitDiffContent($commit)
			];
			$mailBody = $this->view->render($bindings);
			$mailBody = $this->includeStyleSheets($mailBody, 'Resources/StyleSheets/Mail.css');

			// Mail Handling
			$transport = Transport::fromDsn('native://default');
			//$transport = new SendmailTransport(' -t -f' . $this->settings['Configuration']['MailConfiguration']['from']);
			//$transport = new SendmailTransport(' -t');
			$mailer = new Mailer($transport);
			$commitMessageAsArray = explode(chr(10), $commit['message']);
			$subject = 'Commit: ' . $projectData['name_with_namespace'] . ' / ' .
				$branchName . ' ' . $commitMessageAsArray[0];
			$email = (new Email())
				->from($this->settings['Configuration']['MailConfiguration']['from'])
				->subject($subject)
				->html($mailBody);
				
			// Send the message
			$recipients = $utility->getRecipientsForProject($projectData);
			$debugRecipient = [$this->settings['Configuration']['Debug']['debugMailAdresse']];
			$recipients = $this->debugMode ? $debugRecipient : $recipients;

			foreach ($recipients as $mail => $name) {
				if (filter_var($name, FILTER_VALIDATE_EMAIL)) {
					$email->to(new Address($name));
				} else {
					$email->to(new Address($mail, $name));
				}
				$mailer->send($email);
			}

			// log commit id to prevent sending mails twice
			$utility->logCommitId($commit['id']);
		}
	}

	/**
	 * The method downloads the diff file of the given commit and return the content as string.
	 *
	 * https://docs.gitlab.com/ee/api/commits.html
	 *
	 * @param array $commit
	 * @return string
	 */
	protected function crawlCommitDiffContent($commit) {
		// https://gitlab.sgalinski.de/api/v4/projects/medion%2ffirerunner-react-native/repository/commits/<SHA>/diff?private_token=XXXX
		if (!array_key_exists($commit['id'], self::$diffCache)) {
			$diffUrl = $this->settings['Configuration']['Gitlab']['apiUrl'] . '/projects/' . $this->request['project_id'] . '/repository/commits/' . $commit['id'] . '/diff?private_token=' . $this->settings['Configuration']['Gitlab']['authToken'];
			$content = file_get_contents($diffUrl);
			$contentJson = json_decode($content);

			// [old_path] => Classes/Webhook.php
			// [new_path] => Classes/Webhook.php
            // [a_mode] => 100644
            // [b_mode] => 100644
            // [new_file] => 
            // [renamed_file] => 
            // [deleted_file] => 
			// [diff] =>
			$output = '';
			foreach ($contentJson as $diff) {
				$output .= "\n" . 'Old Path: ' . $diff->old_path . "\n";
				$output .= 'New Path: ' . $diff->new_path . "\n";
				$output .= 'Old Mode: ' . $diff->a_mode . "\n";
				$output .= 'New Mode: ' . $diff->b_mode . "\n";
				$output .= $diff->diff . "\n\n--------------------\n\n";
			}

			self::$diffCache[$commit['id']] = $output;

			//error_log(print_r($contentJson, TRUE));
			//die();
		}

		return (self::$diffCache[$commit['id']] !== '') ? self::$diffCache[$commit['id']] : '';
	}

	/**
	 * Method includes stylesheets as inline css to the given markup.
	 *
	 * @param string $markup
	 * @param string $pathToCssFile
	 * @return string
	 */
	protected function includeStyleSheets($markup, $pathToCssFile) {
		$css = file_get_contents($pathToCssFile);

		if ($css) {
			// convert CSS to inline styles for GMail
			$inline = new CssToInlineStyles();
			$markup = $inline->convert($markup, $css);
		}

		return $markup;
	}

	/**
	 * Setter for settings array.
	 *
	 * @param array $settings
	 * @return void
	 */
	public function setSettings(array $settings) {
		$this->settings = $settings;
	}

	/**
	 * Getter for settings array.
	 *
	 * @return array
	 */
	public function getSettings() {
		return $this->settings;
	}

	/**
	 * Setter for request array.
	 *
	 * @param array $request
	 * @return void
	 */
	public function setRequest(array $request) {
		$this->request = $request;
	}

	/**
	 * Getter for request array.
	 *
	 * @return array
	 */
	public function getRequest() {
		return $this->request;
	}
}
